from dataclasses import dataclass
from typing import Tuple, Optional, Literal
import torch.nn as nn


@dataclass
class CNNConfig:
    stack: Tuple[int, ...] = (1,)
    stack_filters: Tuple[int, ...] = (32,)
    kernel_size: Tuple[int, int] = (3, 3)
    type: Literal["cnn"] = "cnn"

    def __post_init__(self):
        if len(self.stack) <= 0:
            raise ValueError("Stack must be positive")
        if any(f <= 0 for f in self.stack_filters):
            raise ValueError("All filter values must be positive")
        if any(v <= 0 for v in self.kernel_size):
            raise ValueError("Kernel values must be positive")
        if len(self.stack) != len(self.stack_filters):
            raise ValueError(
                "Length of stacks and length of stack filters must correspond."
                f" Got {len(self.stack)} != {len(self.stack_filters)}"
            )


@dataclass
class DenseConfig:
    type: Literal["dense"] = "dense"
    dense_units: Tuple[int, ...] = (128, 64)

    def __post_init__(self):
        if any(u <= 0 for u in self.dense_units):
            raise ValueError("dense_units must contain positive values")
        if len(self.dense_units) < 1:
            raise ValueError("dense_units should contain at least an output layer")


@dataclass
class DiresaConfig:
    """
    Global configuration that contains specific architecture configs.
    Can have CNN, Dense, or both.
    """

    input_shape: Tuple[int, ...] = ()
    activation: nn.Module = (
        nn.ReLU()
    )  # Type checker does not agree here that subclass of nn.Relu() is nn.Module for some reason.
    cnn_config: Optional[CNNConfig] = None
    dense_config: Optional[DenseConfig] = None

    def __post_init__(self):
        if len(self.input_shape) == 1:
            if self.cnn_config is not None:
                raise ValueError(
                    "Input shape of 1 is valid only when using dense layers"
                )
        elif len(self.input_shape) == 3:
            if self.cnn_config is None:
                raise ValueError("Input shape of 3 is only valid when using CNN layers")

            # used to verify input shape
            def check_function(x):
                return x % (2 ** len(self.cnn_config.stack)) == 0

            # do not need to check the first dimension as it is depth of input (= channels).
            # Only the 2 last are required to have the right dimensions
            multiple_of_stack_len = list(map(check_function, self.input_shape[1:]))

            if not all(multiple_of_stack_len):
                offenders = [
                    v
                    for v in self.input_shape[1:]
                    if v not in filter(check_function, self.input_shape[1:])
                ]
                raise ValueError(
                    "Dimensions of input shape should be multiple of 2^len(stack)."
                    f" Got {offenders} from {self.input_shape} which are not multiple"
                    f" of 2^{len(self.cnn_config.stack)} (={2**len(self.cnn_config.stack)}) from"
                    f" stack = {self.cnn_config.stack} and input_shape = {self.input_shape}."
                    " This requirement is present because of the pooling layers in the CNN layers"
                    " which divide the dimensions by 2 at each stack."
                )
        else:
            raise ValueError(
                "Must provide valid input_shape. Rank 3 with Conv2D layers "
                "with shape (Height, Width, Depth) and Dims should be a multiple of 2^len(stack)"
                f"; Rank 1 if only Dense layer. Got {self.input_shape}"
            )


class DiresaSetup:
    """
    Factory class providing ``create_config`` as factory method to create configuration from user parameters
    """

    @staticmethod
    def create_config(
        input_shape: Tuple[int, ...] = (),
        # CNN parameters
        stack: Optional[Tuple[int, ...]] = None,
        stack_filters: Optional[Tuple[int, ...]] = None,
        kernel_size: Tuple[int, int] = (3, 3),
        # Dense parameters
        dense_units: Optional[Tuple[int, ...]] = None,
        # Global parameters
        activation: nn.Module = nn.ReLU(),
    ) -> DiresaConfig:
        """
        Factory Method to build a DiresaConfig from hyperparameters

        :param input_shape: is the shape of input features. Rank 2 with Conv2D layers, first 2 dims should be a multiple of 2^len(stack); rank 1 if only Dense layers
        :param stack: are the number of Conv2D layers in each block
        :param stack_filters: number of filters in a block
        :param kernel_size: kernel size for convultion
        :param dense_units: tuple which describes the width of each full connected feedforward layers. First layer describes the input size, the last layer defines the output layer which is the size of latent space. E.g. ``(3, 16, 32, 64)`` is a FFN with one input layer of size ``3``, two hidden layers ``(16, 32)`` and last layer which defines latent space of size ``64``. When using dense_units in combination with CNN layers an additional layer is added to match CNN output to FFN input. The input layer to the FNN is effectively already computed by the model.
        :param activation: activation function used through the network.

        :return: ``DiresaConfig`` which is the global configuration object containing the appropriate sub-configs


        :raises ValueError: in case of an invalid configuration
        """
        # Check what type of arch we deal with
        has_cnn_params = stack is not None and stack_filters is not None
        has_dense_params = dense_units is not None

        if not has_cnn_params and not has_dense_params:
            raise ValueError(
                "Must provide either CNN parameters (stack + stack_filters) "
                "or dense parameters (dense_units) or both"
            )

        # CNN config if parameters provided
        cnn_config = None
        if has_cnn_params:
            cnn_config = CNNConfig(
                stack=stack,
                stack_filters=stack_filters,
                kernel_size=kernel_size,
            )

        # Dense config if parameters provided
        dense_config = None
        if has_dense_params:
            dense_config = DenseConfig(dense_units=dense_units)

        # Create global config
        return DiresaConfig(
            input_shape=input_shape,
            activation=activation,
            cnn_config=cnn_config,
            dense_config=dense_config,
        )



import torch
import torch.nn as nn


class LatentCovLoss(nn.Module):
    """
    Computes Covariance Loss on input batch.
    """

    # TODO: Simulated Annealing. Not present by default in pytorch. Needs custom implementation
    # With staged training might not really need it anymore as parameters are tuned independently

    def __init__(self):
        super().__init__()

    def forward(self, latent, _):
        """
        :param latent: is the latent space representation of the current batch
        :return: covariance loss
        """
        cov = torch.cov(torch.transpose(latent, 0, 1), correction=1)
        cov_square = cov * cov  # elem wise mult
        # number of covariance entries (need to subtract 1 to not take into account diagonal variance metrics)
        nbr_of_cov = latent.shape[-1] * (latent.shape[-1] - 1)
        cov_loss = (torch.sum(cov_square) - torch.trace(cov_square)) / float(nbr_of_cov)
        return cov_loss


class MAEDistLoss(nn.Module):
    """
    Absolute Error between original and latent distances
    """

    def __init__(self):
        super().__init__()

    def forward(self, distances, _):
        """
        :param distances: batch of original and latent distances between twins
        :return: mean of absolute errors
        """
        ae = torch.abs(distances[:, 0] - distances[:, 1])
        return torch.mean(ae)


class MALEDistLoss(nn.Module):
    """
    Absolute Error between logarithm of original and latent distances
    """

    def __init__(self, factor=1.):
        """
        :param factor: distance multiplication factor
        """
        super().__init__()
        self.factor = factor

    def forward(self, distances, _):
        """
        :param distances: batch of original and latent distances between twins
        :return: mean of absolute logarithmic errors
        """
        ale = torch.abs(torch.log1p(self.factor * distances[:, 0]) - torch.log1p(self.factor * distances[:, 1]))
        return torch.mean(ale)


class MAPEDistLoss(nn.Module):
    """
    Absolute Percentage Error between original and latent distances
    """

    def __init__(self):
        super().__init__()

    def forward(self, distances, _):
        """
        :param distances: batch of original and latent distances between twins
        :return: mean of absolute percentage errors
        """
        epsilon = 1e-8
        ape = torch.abs(
            (distances[:, 0] - distances[:, 1]) / (distances[:, 0] + epsilon)
        )
        return torch.mean(ape)


class MSEDistLoss(nn.Module):
    """
    Squared Error between original and latent distances
    """

    def __init__(self):
        super().__init__()

    def forward(self, distances, _):
        """
        :param distances: batch of original and latent distances between twins
        :return: mean of squared errors
        """
        se = torch.square(distances[:, 0] - distances[:, 1])
        return torch.mean(se)


class MSLEDistLoss(nn.Module):
    """
    Squared Error between logarithm of original and latent distances
    """

    def __init__(self, factor=1.):
        """
        :param factor: distance multiplication factor
        """
        super().__init__()
        self.factor = factor

    def forward(self, distances, _):
        """
        :param distances: batch of original and latent distances between twins
        :return: mean of squared logarithmic errors
        """
        sle = torch.square(torch.log1p(self.factor * distances[:, 0]) - torch.log1p(self.factor * distances[:, 1]))
        return torch.mean(sle)


class CorrDistLoss(nn.Module):
    """
    Correlation loss between original and latent distances
    """

    def __init__(self):
        super().__init__()

    def forward(self, distances, _):
        """
        :param distances: batch of original and latent distances between twins
        :return: 1 - correlation coefficient
        """
        cov = torch.cov(torch.transpose(distances, 0, 1), correction=1)
        cov_sqrt = torch.sqrt(torch.abs(cov))
        corr = 1 - cov[0, 1] / (cov_sqrt[0, 0] * cov_sqrt[1, 1])
        return 1 - corr


class CorrLogDistLoss(nn.Module):
    """
    Correlation loss between  logarithm of original and latent distances
    """

    def __init__(self, factor=1.):
        """
        :param factor: distance multiplication factor
        """
        super().__init__()
        self.factor = factor

    def forward(self, distances, _):
        """
        :param distances: batch of original and latent distances between twins
        :return: 1 - correlation coefficient (of logarithmic distances)
        """
        cov = torch.cov(torch.transpose(torch.log1p(self.factor * distances), 0, 1), correction=1)
        cov_sqrt = torch.sqrt(torch.abs(cov))
        corr = 1 - cov[0, 1] / (cov_sqrt[0, 0] * cov_sqrt[1, 1])
        return 1 - corr



import torch
from typing import Optional, Dict
import logging
from torch.utils.data import DataLoader
from diresa_torch.arch.models import Diresa
from diresa_torch.utils.utils import _r2_score, _set_components_to_mean
from operator import add


def __compute_losses(outputs, targets, criteria, loss_weights):
    """
    Helper function to compute losses given outputs, targets, criteria and weights.

    :param outputs: Model outputs (tuple of 3 elements: reconstructed, latent, distance)
    :param targets: Target values (tuple of 3 elements: data, None, None)
    :param criteria: List of loss functions [ReconstructionLoss, CovarianceLoss, DistanceLoss]
    :param loss_weights: Weighting factor for the different losses
    :return: Tuple of (individual_losses, total_weighted_loss)
    """
    individual_losses = [c(o, t) for c, o, t in zip(criteria, outputs, targets)]
    weighted_losses = [w * l for w, l in zip(loss_weights, individual_losses)]
    total_weighted_loss = torch.stack(weighted_losses).sum()

    return individual_losses, total_weighted_loss


def __set_non_trainable(model):
    for param in model.parameters():
        param.requires_grad = False


def __loss_string_repr(criteria, suffix=""):
    """
    Helper functions producing string criteria of
    losses. String repr is used to log each loss with an
    understandable name.
    """
    # Ordering of loss output values
    # Use class name without the last () as name for the loss
    loss_names = [f"{c}"[:-2] for c in criteria] + ["WeightedLoss"]

    # Criteria list can change depending on training mode.
    # When running normal criteria[0] is Reconstruction loss (and len(criteria == 3))
    # During staged training, when encoder is trained
    # criteria[0] is LatentCovLoss (and len(criteria == 2)).
    # When decoder is trained criteria[0] is Recon (and len(criteria == 1))
    if len(criteria) == 3 or len(criteria) == 1:
        loss_names[0] = "Recon" + loss_names[0]

    # add "train" suffix for output
    loss_names = list(map(lambda x: x + suffix, loss_names))
    return loss_names


def __evaluate(
    produce_output: callable,
    produce_target: callable,
    test_loader: DataLoader,
    criteria: list,
    device: torch.device,  # device is still required here as this function does not have access to model
    loss_weights: list,
    input_filter: callable,
    target_filter: callable,
    loss_suffix: str,
) -> Dict[str, float]:
    """
    Evaluate DIRESA (does not track gradient) by computing all three losses (reconstruction, covariance, distance) with help of the ``produce_output`` and ``produce_input`` functions. Those functions are provided as lambdas which makes it easier to match the outputs, targets and criteria together for evaluation purposes.

    :param produce_output: callable function producing outputs from model. Takes as input batch data.
    :param produce_target: callable function producing target values for criterion. Takes as input batch data.
    :param test_loader: Test data loader
    :param criteria: List of loss functions, depends on what part is being evaluated.
    :param device: Device to evaluate on
    :param loss_weights: Weighting factor for the different losses (in order [Reconstruction, Covariance, Distance])
    :param loss_suffix: Appends a suffix to the loss as __eval can be used for validation during training or evaluation (when passing in test set)
    :param input_filter: Function used to filter input data
    :param target_filter: Function used to filter target data
    :return: Dictionary with average losses: individual losses + weighted total loss
    """
    total_losses = [0.0] * (len(criteria) + 1)
    num_batches = 0

    loss_names = __loss_string_repr(criteria, loss_suffix)

    with torch.no_grad():
        for _, data in enumerate(test_loader):
            data = data.to(device)

            input_data = input_filter(data)
            target_data = target_filter(data)
            targets = produce_target(target_data)
            outputs = produce_output(input_data)

            # outputs_losses, loss = _compute_losses(outputs, target, criteria, loss_weights)
            outputs_losses, loss = __compute_losses(
                outputs, targets, criteria, loss_weights
            )

            # Accumulate losses
            all_losses = outputs_losses + [loss]
            total_losses = list(
                map(add, total_losses, [loss.item() for loss in all_losses])
            )
            num_batches += 1

    avg_losses = [loss / num_batches for loss in total_losses]

    result = {name: avg_loss for name, avg_loss in zip(loss_names, avg_losses)}

    return result


def train_diresa(
    model: Diresa,
    train_loader: DataLoader,
    criteria: list,
    optimizer: torch.optim.Optimizer,
    num_epochs: int = 10,
    val_loader: Optional[DataLoader] = None,
    loss_weights: list = [1.0, 1.0, 1.0],
    staged_training: bool = False,
    input_filter: callable = lambda data: data,
    target_filter: callable = lambda data: data,
    callbacks: Optional[list] = None,  # Not IMPL at the moment
) -> dict[str, dict[str, list[float]]]:
    """
    Trains `model`. Needs to provide multiple loss function in order to train de different parts of the model.
    CovarianceLoss and DistanceLoss are used to produce an interpretable latent space
    while ReconstructionLoss is used to produce an output.

    :param model: The model to train
    :param train_loader: Training data loader
    :param criteria: List of Loss function. With order [ReconstructionLoss, CovarianceLoss, DistanceLoss]
    :param optimizer: Optimizer
    :param num_epochs: Number of epochs
    :param val_loader: Optional validation loader
    :param loss_weights: Weighting factor for the different losses. With order [ReconstructionLoss, CovarianceLoss, DistanceLoss]
    :param staged_training: If set to True will train the encoder and the decoder separately for `num_epochs` each.
    :param input_filter: Function used to filter input data, default is no filtering
    :param target_filter: Function used to filter target data, default is no filtering
    :param callbacks: Optional list of callback functions
    :return: Dict with training (losses, metrics) and validation (if val_loader is provided).
    """
    assert (
        callbacks is None
    ), "Callbacks are not tested at the moment. Remove param or set to None"

    def __train_for_epochs(
        produce_output: callable,
        produce_target: callable,
        criteria,
        loss_weights,
        device,
        prepend_log="DIRESA",
    ):
        """
        Nested function for training loop. Factors out common functionalities for training,
        while providing custom information about what loss to train for used to differentiate
        between staged training and full training.

        :param produce_output: callable function producing outputs from model. Takes as input batch data.
        :param produce_target: callable function producing target values for criterion. Takes as input batch data.
        :param criteria: List of loss functions
        :param loss_weights: weights for each loss function
        :param device: hardware device used.
        :param prepend_log: String to prepend to logging output
        """
        assert len(criteria) == len(
            loss_weights
        ), "Number of criteria and their associated weights does not match"

        loss_names = __loss_string_repr(criteria, "_train")
        history = {name: [] for name in loss_names}

        for epoch in range(num_epochs):
            # each criterion loss + combined weighted loss
            epoch_loss = [0.0] * (len(criteria) + 1)
            num_batches = 0

            for _, data in enumerate(train_loader):
                data = data.to(device)
                input_data = input_filter(data)
                target_data = target_filter(data)
                target = produce_target(target_data)
                outputs = produce_output(input_data)

                optimizer.zero_grad()
                outputs_losses, loss = __compute_losses(
                    outputs, target, criteria, loss_weights
                )

                # accumulates gradient in each tensor -> Backprop
                # backpropagated loss in weighted sum of each loss.
                loss.backward()

                optimizer.step()

                # add weighted loss to final losses
                all_losses = outputs_losses + [loss]
                epoch_loss = list(
                    map(add, epoch_loss, [loss.item() for loss in all_losses])
                )
                num_batches += 1

                if callbacks:
                    raise NotImplementedError
                    # for callback in callbacks:
                    #     callback(epoch, batch_idx, loss.item())

            avg_loss = list(map(lambda loss: loss / num_batches, epoch_loss))

            for name, loss in zip(loss_names, avg_loss):
                history[name].append(loss)

            # val loader is defined in exterior function
            if val_loader:
                val_dict = __evaluate(
                    produce_output,
                    produce_target,
                    val_loader,
                    criteria,
                    device,
                    loss_weights,
                    input_filter,
                    target_filter,
                    loss_suffix="_val",
                )
                for name, loss in val_dict.items():
                    if name in history:
                        history[name].append(loss)
                    else:
                        history[name] = [loss]

            # print out last entry in history for each epoch
            log_str = ", ".join(
                [f"{name}: {values[-1]:.4e}" for name, values in history.items()]
            )
            logging.info(f"{prepend_log}: Epoch {epoch + 1}/{num_epochs} - {log_str}")

        return history

    # End of nested function

    # takes the device onto which the first tensor is registered
    device = next(model.parameters()).device

    if staged_training:
        # train encoder, cov and dist loss
        hist_encoder = __train_for_epochs(
            produce_output=lambda data: model._encode_with_distance(data),
            produce_target=lambda _: (None, None),
            criteria=criteria[1:],  # cov and dist criteria
            loss_weights=loss_weights[1:],  # cov and dist weights
            device=device,
            prepend_log="Encoder",
        )

        # freeze encoder weights
        __set_non_trainable(model.base_encoder)

        # train decoder, only rec loss
        hist_decoder = __train_for_epochs(
            produce_output=lambda data: model.base_decoder(model.base_encoder(data)),
            produce_target=lambda data: data,
            criteria=criteria[:1],
            loss_weights=loss_weights[:1],
            device=device,
            prepend_log="Decoder",
        )

        hist = {"Encoder": hist_encoder, "Decoder": hist_decoder}
        return hist

    else:
        # data is produced by forward pass of model.
        hist_diresa = __train_for_epochs(
            lambda data: model(data),
            lambda data: (data, None, None),
            criteria,
            loss_weights,
            device=device,
            prepend_log="Encoder_Decoder",
        )

        # To keep consistency with staged training where encoder
        # and decoder losses are accessed via a key, the same is done
        # when all weights are trained simultaneously.
        hist = {"Encoder_Decoder": hist_diresa}
        return hist


def evaluate_diresa(
    model: Diresa,
    test_loader: DataLoader,
    criteria: list,
    loss_weights: list = [1.0, 1.0, 1.0],
    input_filter: callable = lambda data: data,
    target_filter: callable = lambda data: data,
) -> Dict[str, float]:
    """
    Evaluates `model` using `test_loader`

    :param model: The model to evaluate
    :param test_loader: Test data loader
    :param criteria: List of loss functions [ReconstructionLoss, CovarianceLoss, DistanceLoss]
    :param loss_weights: Weighting factor for the different losses
    :param input_filter: Function used to filter input data, default is no filtering
    :param target_filter: Function used to filter target data, default is no filtering
    :return: Dictionary with averaged losses: individual criterion loss + weighted total loss
    """
    assert (
        len(criteria) == 3
    ), "Need to provide 3 criteria for DIRESA evaluation, namely [ReconstructionLoss, CovarianceLoss, DistanceLoss]"

    device: torch.device = next(model.parameters()).device

    eval_dict = __evaluate(
        # model.forward(data) produces (reconstructed, latent, dist)
        lambda data: model.forward(data),
        lambda data: (data, None, None),
        test_loader=test_loader,
        device=device,
        criteria=criteria,
        loss_weights=loss_weights,
        input_filter=input_filter,
        target_filter=target_filter,
        loss_suffix="_eval",
    )
    log_str = ", ".join([f"{name}: {value:.4e}" for name, value in eval_dict.items()])
    logging.info(log_str)
    return eval_dict


def predict_diresa(
    model: Diresa,
    data_loader: DataLoader,
    input_filter: callable = lambda data: data,
) -> torch.Tensor:
    """
    predict_diresa is the reconstructed dataset from `data_loader` passed through `model`.
    Provides faster inference as distance and covariance are not computed for inference.

    :param model: model to use to produce a prediction
    :param data_loader: data to be reconstructed
    :param input_filter: function used to filter input data, default is no filtering
    :return: prediction
    """

    device = next(model.parameters()).device
    predictions = []

    with torch.no_grad():
        for _, data in enumerate(data_loader):
            data = data.to(device)
            input_data = input_filter(data)
            outputs = model.fast_eval(input_data)
            predictions.append(outputs.cpu())

    return torch.cat(predictions, dim=0)


def order_diresa(
    model: Diresa,
    data_loader: DataLoader,
    cumul=False,
    input_filter: callable = lambda data: data,
    target_filter: callable = lambda data: data,
) -> list:
    """
    Sets ordering of the OrderingLayer.
    Limitations: assumes a flat latent space (rank of latent is 2).
    If `cumul` is set to true it iteratively selects next component based on
    additional combined explanatory power with previously selected components, if it is set
    to false it only sorts based on the R² of each latent dimension separately.

    :param model: The model on which to produce the ordering
    :param data_loader: The data_loader from which to produce the ordering
    :param cumul: If `False` only sorts based on single-dimension R²
    :param input_filter: Function used to filter input data, default is no filtering
    :param target_filter: Function used to filter target data, default is no filtering
    :return: (cumulative) R² scores of latent components
    """

    logging.info(f"Batch size for ordering is {data_loader.batch_size}")

    device = next(model.parameters()).device
    first_r2_scores = []

    with torch.no_grad():
        R2_full = []
        # produce R2 per batch
        for _, data in enumerate(data_loader):
            data = data.to(device)

            # 1. Produce latent dataset
            input_data = input_filter(data)
            target_data = target_filter(data)
            latent = model.base_encoder(input_data)
            assert len(latent.shape) == 2, "Latent space is not flattened"
            R2_full.append(_r2_score(target_data, model.base_decoder(latent)))

            # 2. Produce l latent samples for which every latent dimensions is averaged except the l-th one
            averaged = map(
                lambda i: _set_components_to_mean(latent, i),
                range(latent.shape[1]),
            )

            # 3. Produce l decoded samples from l latent samples
            decoded = map(lambda latent: model.base_decoder(latent), averaged)

            # 4. Compute R2 by comparing l decoded with original x
            r2 = list(map(lambda pred: _r2_score(target_data, pred), decoded))

            first_r2_scores.append(r2)

        first_r2_scores = torch.tensor(first_r2_scores, device=device).mean(dim=0)
        logging.info(f"Total R² score is: {torch.tensor(R2_full, device=device).mean(dim=0)}")

        first_ordering = torch.argsort(first_r2_scores, descending=True)

        if not cumul:
            ordering = first_ordering
            r2_scores = first_r2_scores[ordering].tolist()

        else:
            # Full cumulative iterative selection
            ordering = [first_ordering[0].item()]
            # Start the selection with the best single-R2 dimension. Ordering keeps track of selected latent components in order
            # Remaining will be a set of all the latent indices, except for the ones we already chose
            # So it will hold the possible candidates
            remaining = set(range(first_r2_scores.shape[0])) - set(ordering)
            # Now we compute the r2 score of the first latent, to see what we are improving upon
            r2_scores = [first_r2_scores[ordering[0]].item()]

            # Now we want to loop n-1 times, until all dimensions are fixed.
            while remaining:
                candidate_scores = {}  # Dictionary with key = candidate index, value = combined R2 scores if we would add this one next

                for cand in remaining:
                    r2_subscores = []
                    for _, data in enumerate(data_loader):
                        data = data.to(device)

                        input_data = input_filter(data)
                        target_data = target_filter(data)
                        latent = model.base_encoder(input_data)

                        # Keep only ordering + this candidate active
                        active_dims = list(set(ordering) | {cand})
                        # Now we isolate the combined effect of the excluded dimensions:
                        mod_latent = _set_components_to_mean(latent, active_dims)
                        decoded = model.base_decoder(mod_latent)
                        r2_subscores.append(_r2_score(target_data, decoded))

                    cand_score = torch.tensor(r2_subscores,
                                              device=device).mean().item()
                    candidate_scores[cand] = cand_score

                # Select the next best latent
                # When comparing tuples (index, score), sort by the score, so element 1 
                next_best = max(candidate_scores.items(), key=lambda x: x[1])
                ordering.append(next_best[0])
                remaining.remove(next_best[0])
                r2_scores.append(next_best[1])

            ordering = torch.tensor(ordering, device=device)

        model.ordering_layer.order = ordering
        logging.info(f"Ordered R² scores are: {r2_scores}")

        return r2_scores



import torch
from diresa_torch.arch.models import Diresa


def save_diresa(model: Diresa, path: str):
    """
    Saves Diresa model parameters to disk

    :param model: Diresa model to save
    :param path: path/file to save model paramaters
    """
    torch.save(model.state_dict(), path)


def load_diresa(model: Diresa, path: str) -> Diresa:
    """
    Loads Diresa model parameters from disk

    :param model: Diresa model to load
    :param path: path/file to load model parameters
    :return: Diresa model
    """
    diresa_state = torch.load(path)
    if 'ordering_layer._order' in diresa_state:
        model.ordering_layer.order = diresa_state['ordering_layer._order']
    model.load_state_dict(diresa_state)
    return model


def _r2_score(y: torch.Tensor, y_pred: torch.Tensor):
    """
    :param y: original dataset
    :param y_pred: predicted dataset
    :return: R2 score between y and y_pred
    """
    error = torch.sum(torch.square(y - y_pred))
    var = torch.sum(torch.square(y - torch.mean(y, dim=0)))
    r2 = 1.0 - error / var
    return r2.item()  # Convert to Python scalar


def _set_components_to_mean(latent: torch.Tensor, retain_idxs: list):
    """
    Set all latent components to mean except the ones in retain_idxs.
    retain_idxs may be a single int or a list of ints, corresponding to the indexes of these dimensions.

    :param latent: latent dataset
    :param retain_idxs: list of indexes which will be retained and thus not set to their mean value
    :return: a new latent dataset with the dimensions which were not in retain_idxs set to their mean value
    """
    if isinstance(retain_idxs, int):
        retain_idxs = [retain_idxs]

    with torch.no_grad():
        mean_values = latent.mean(dim=0, keepdim=True)

        mask = torch.ones_like(latent, dtype=torch.bool)
        mask[:, retain_idxs] = False  # these dims keep original values

        result = torch.where(mask, mean_values, latent)
    return result
